@@ -2,11 +2,11 @@
 #   Pyrex - Types
 #
 
+from Cython.Utils import UtilityCode
 import StringEncoding
 import Naming
 import copy
 
-
 class BaseType:
     #
     #  Base class for all Pyrex types including pseudo-types.
@@ -29,6 +29,7 @@ class PyrexType(BaseType):
     #  is_extension_type     boolean     Is a Python extension type
     #  is_numeric            boolean     Is a C numeric type
     #  is_int                boolean     Is a C integer type
+    #  is_longlong           boolean     Is a long long or unsigned long long.
     #  is_float              boolean     Is a C floating point type
     #  is_void               boolean     Is the C void type
     #  is_array              boolean     Is a C array type
@@ -36,6 +37,7 @@ class PyrexType(BaseType):
     #  is_null_ptr           boolean     Is the type of NULL
     #  is_cfunction          boolean     Is a C function type
     #  is_struct_or_union    boolean     Is a C struct or union type
+    #  is_struct             boolean     Is a C struct type
     #  is_enum               boolean     Is a C enum type
     #  is_typedef            boolean     Is a typedef type
     #  is_string             boolean     Is a C char * type
@@ -79,6 +81,7 @@ class PyrexType(BaseType):
     is_builtin_type = 0
     is_numeric = 0
     is_int = 0
+    is_longlong = 0
     is_float = 0
     is_void = 0
     is_array = 0
@@ -86,6 +89,7 @@ class PyrexType(BaseType):
     is_null_ptr = 0
     is_cfunction = 0
     is_struct_or_union = 0
+    is_struct = 0
     is_enum = 0
     is_typedef = 0
     is_string = 0
@@ -97,6 +101,7 @@ class PyrexType(BaseType):
     default_value = ""
     parsetuple_format = ""
     pymemberdef_typecode = None
+    typestring = None
     
     def resolve(self):
         # If a typedef, returns the base type.
@@ -136,6 +141,9 @@ class PyrexType(BaseType):
         # a struct whose attributes are not defined, etc.
         return 1
 
+    def is_simple_buffer_dtype(self):
+        return (self.is_int or self.is_float or self.is_pyobject or
+                self.is_extension_type or self.is_ptr)
 
 class CTypedefType(BaseType):
     #
@@ -195,20 +203,24 @@ class BufferType(BaseType):
     #  lookups to the base type. ANYTHING NOT DEFINED
     #  HERE IS DELEGATED!
     
-    # dtype         PyrexType
-    # ndim          int
-    # mode          str
-    # is_buffer     boolean
-    # writable      boolean
+    # dtype            PyrexType
+    # ndim             int
+    # mode             str
+    # negative_indices bool
+    # cast             bool
+    # is_buffer        bool
+    # writable         bool
 
     is_buffer = 1
     writable = True
-    def __init__(self, base, dtype, ndim, mode):
+    def __init__(self, base, dtype, ndim, mode, negative_indices, cast):
         self.base = base
         self.dtype = dtype
         self.ndim = ndim
         self.buffer_ptr_type = CPtrType(dtype)
         self.mode = mode
+        self.negative_indices = negative_indices
+        self.cast = cast
     
     def as_argument_type(self):
         return self
@@ -274,7 +286,6 @@ class BuiltinObjectType(PyObjectType):
         return "<%s>"% self.cname
         
     def assignable_from(self, src_type):
-
         if isinstance(src_type, BuiltinObjectType):
             return src_type.name == self.name
         else:
@@ -290,7 +301,16 @@ class BuiltinObjectType(PyObjectType):
         return type.is_pyobject and self.assignable_from(type)
         
     def type_test_code(self, arg):
-        return 'likely(Py%s_CheckExact(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (self.name[0].upper() + self.name[1:], arg, arg, self.name, arg)
+        type_name = self.name
+        if type_name == 'str':
+            type_name = 'String'
+        elif type_name == 'set':
+            type_name = 'AnySet'
+        elif type_name == 'frozenset':
+            type_name = 'FrozenSet'
+        else:
+            type_name = type_name.capitalize()
+        return 'likely(Py%s_CheckExact(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (type_name, arg, arg, self.name, arg)
 
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
@@ -397,7 +417,10 @@ class CType(PyrexType):
     from_py_function = None
     exception_value = None
     exception_check = 1
-    
+
+    def create_convert_utility_code(self, env):
+        return True
+        
     def error_condition(self, result_code):
         conds = []
         if self.is_string:
@@ -467,6 +490,8 @@ class CNumericType(CType):
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
         base = public_decl(self.sign_and_name(), dll_linkage)
+        if for_display and self.is_longlong:
+            base = base.replace('PY_LONG_LONG', 'long long')
         return self.base_declaration_code(base,  entity_code)
 
 
@@ -551,14 +576,16 @@ class CULongType(CUIntType):
     from_py_function = "PyInt_AsUnsignedLongMask"
 
 
-class CLongLongType(CUIntType):
+class CLongLongType(CIntType):
 
+    is_longlong = 1
     to_py_function = "PyLong_FromLongLong"
     from_py_function = "__pyx_PyInt_AsLongLong"
 
 
 class CULongLongType(CUIntType):
 
+    is_longlong = 1
     to_py_function = "PyLong_FromUnsignedLongLong"
     from_py_function = "__pyx_PyInt_AsUnsignedLongLong"
 
@@ -728,7 +755,7 @@ class CFuncType(CType):
             return 1
         if not other_type.is_cfunction:
             return 0
-        if not self.is_overridable and other_type.is_overridable:
+        if self.is_overridable != other_type.is_overridable:
             return 0
         nargs = len(self.args)
         if nargs != len(other_type.args):
@@ -842,6 +869,8 @@ class CFuncType(CType):
         for arg in self.args[:len(self.args)-self.optional_arg_count]:
             arg_decl_list.append(
                 arg.type.declaration_code("", for_display, pyrex = pyrex))
+        if self.is_overridable:
+            arg_decl_list.append("int %s" % Naming.skip_dispatch_cname)
         if self.optional_arg_count:
             arg_decl_list.append(self.op_arg_struct.declaration_code(Naming.optional_args_cname))
         if self.has_varargs:
@@ -918,6 +947,50 @@ class CStructOrUnionType(CType):
         self.kind = kind
         self.scope = scope
         self.typedef_flag = typedef_flag
+        self.is_struct = kind == 'struct'
+        if self.is_struct:
+            self.to_py_function = "%s_to_py_%s" % (Naming.convert_func_prefix, self.cname)
+        self.exception_check = True
+        self._convert_code = None
+        
+    def create_convert_utility_code(self, env):
+        if env.outer_scope is None:
+            return False
+        if self._convert_code is None:
+            import Code
+            code = Code.CCodeWriter()
+            header = "static PyObject* %s(%s)" % (self.to_py_function, self.declaration_code('s'))
+            code.putln("%s {" % header)
+            code.putln("PyObject* res;")
+            code.putln("PyObject* member;")
+            code.putln("res = PyDict_New(); if (res == NULL) return NULL;")
+            for member in self.scope.var_entries:
+                if member.type.to_py_function and member.type.create_convert_utility_code(env):
+                    interned_name = env.get_string_const(member.name, identifier=True)
+                    env.add_py_string(interned_name)
+                    code.putln("member = %s(s.%s); if (member == NULL) goto bad;" % (
+                                                member.type.to_py_function, member.cname))
+                    code.putln("if (PyDict_SetItem(res, %s, member) < 0) goto bad;" % interned_name.pystring_cname)
+                    code.putln("Py_DECREF(member);")
+                else:
+                    self.to_py_function = None
+                    return False
+            code.putln("return res;")
+            code.putln("bad:")
+            code.putln("Py_XDECREF(member);")
+            code.putln("Py_DECREF(res);")
+            code.putln("return NULL;")
+            code.putln("}")
+            proto = header + ";"
+            # This is a bit of a hack, we need a forward declaration
+            # due to the way things are ordered in the module...
+            entry = env.lookup(self.name)
+            if entry.visibility != 'extern':
+                proto = self.declaration_code('') + ';\n' + proto
+            self._convert_code = UtilityCode(proto=proto, impl=code.buffer.getvalue())
+        
+        env.use_utility_code(self._convert_code)
+        return True
         
     def __repr__(self):
         return "<CStructOrUnionType %s %s%s>" % (self.name, self.cname,
@@ -951,6 +1024,11 @@ class CStructOrUnionType(CType):
     def attributes_known(self):
         return self.is_complete()
 
+    def can_be_complex(self):
+        # Does the struct consist of exactly two floats?
+        fields = self.scope.var_entries
+        return len(fields) == 2 and fields[0].type.is_float and fields[1].type.is_float
+
 
 class CEnumType(CType):
     #  name           string
@@ -1036,6 +1114,17 @@ class CCharPtrType(CStringType, CPtrType):
         CPtrType.__init__(self, c_char_type)
 
 
+class UnspecifiedType(PyrexType):
+    # Used as a placeholder until the type can be determined.
+        
+    def declaration_code(self, entity_code, 
+            for_display = 0, dll_linkage = None, pyrex = 0):
+        return "<unspecified>"
+    
+    def same_as_resolved_type(self, other_type):
+        return False
+        
+
 class ErrorType(PyrexType):
     # Used to prevent propagation of error messages.
     
@@ -1046,6 +1135,9 @@ class ErrorType(PyrexType):
     from_py_function = "dummy"
     typestring = None
     
+    def create_convert_utility_code(self, env):
+        return True
+    
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
         return "<error>"
@@ -1116,6 +1208,7 @@ c_py_buffer_type = CStructOrUnionType("Py_buffer", "struct", None, 1, "Py_buffer
 c_py_buffer_ptr_type = CPtrType(c_py_buffer_type)
 
 error_type =    ErrorType()
+unspecified_type = UnspecifiedType()
 
 lowest_float_rank = 6
 
@@ -1142,6 +1235,10 @@ sign_and_rank_to_type = {
     (1, 6): c_float_type, 
     (1, 7): c_double_type,
     (1, 8): c_longdouble_type,
+# In case we're mixing unsigned ints and floats...
+    (0, 6): c_float_type, 
+    (0, 7): c_double_type,
+    (0, 8): c_longdouble_type,
 }
 
 modifiers_and_name_to_type = {
@@ -1169,28 +1266,55 @@ modifiers_and_name_to_type = {
     (2, 1, "int"): c_slong_type,
     (2, 2, "int"): c_slonglong_type,
     (2, 0, "Py_ssize_t"): c_py_ssize_t_type,
+    
+    (1, 0, "long"): c_long_type,
+    (1, 0, "short"): c_short_type,
+    (1, 0, "longlong"): c_longlong_type,
+    (1, 0, "bint"): c_bint_type,
 }
 
 def widest_numeric_type(type1, type2):
     # Given two numeric types, return the narrowest type
     # encompassing both of them.
     if type1.is_enum and type2.is_enum:
-        widest_type = c_int_type
-    elif type2.rank > type1.rank:
-        widest_type = type2
+        return c_int_type
+    elif type1 is type2:
+        return type1
+    elif (type1.signed and type2.signed) or (not type1.signed and not type2.signed):
+        if type2.rank > type1.rank:
+            return type2
+        else:
+            return type1
     else:
-        widest_type = type1
+        return sign_and_rank_to_type[min(type1.signed, type2.signed), max(type1.rank, type2.rank)]
     return widest_type
 
 def simple_c_type(signed, longness, name):
     # Find type descriptor for simple type given name and modifiers.
     # Returns None if arguments don't make sense.
     return modifiers_and_name_to_type.get((signed, longness, name))
+    
+def parse_basic_type(name):
+    base = None
+    if name.startswith('p_'):
+        base = parse_basic_type(name[2:])
+    elif name.startswith('p'):
+        base = parse_basic_type(name[1:])
+    elif name.endswith('*'):
+        base = parse_basic_type(name[:-1])
+    if base:
+        return CPtrType(base)
+    elif name.startswith('u'):
+        return simple_c_type(0, 0, name[1:])
+    else:
+        return simple_c_type(1, 0, name)
 
 def c_array_type(base_type, size):
     # Construct a C array type.
     if base_type is c_char_type:
         return CCharArrayType(size)
+    elif base_type is error_type:
+        return error_type
     else:
         return CArrayType(base_type, size)
 
@@ -1198,8 +1322,23 @@ def c_ptr_type(base_type):
     # Construct a C pointer type.
     if base_type is c_char_type:
         return c_char_ptr_type
+    elif base_type is error_type:
+        return error_type
     else:
         return CPtrType(base_type)
+        
+def Node_to_type(node, env):
+    from ExprNodes import NameNode, AttributeNode, StringNode, error
+    if isinstance(node, StringNode):
+        node = NameNode(node.pos, name=node.value)
+    if isinstance(node, NameNode) and node.name in rank_to_type_name:
+        return simple_c_type(1, 0, node.name)
+    elif isinstance(node, (AttributeNode, NameNode)):
+        node.analyze_types(env)
+        if not node.entry.is_type:
+            pass
+    else:
+        error(node.pos, "Bad type")
 
 def public_decl(base, dll_linkage):
     if dll_linkage:
